package org.pwte.example.service;

import java.math.BigDecimal;

import javax.ejb.Stateless;
import javax.persistence.*;



import org.pwte.example.domain.AbstractCustomer;
import org.pwte.example.domain.LineItem;
import org.pwte.example.domain.LineItemId;
import org.pwte.example.domain.Order;
import org.pwte.example.domain.Product;
import org.pwte.example.exception.CustomerDoesNotExistException;
import org.pwte.example.exception.GeneralPersistenceException;
import org.pwte.example.exception.InvalidQuantityException;
import org.pwte.example.exception.NoLineItemsException;
import org.pwte.example.exception.OrderAlreadyOpenException;
import org.pwte.example.exception.OrderNotOpenException;
import org.pwte.example.exception.ProductDoesNotExistException;

@Stateless
public class CustomerOrderServicesImpl implements CustomerOrderServices {

	@PersistenceContext(unitName="pie-db-JAVA-EE")
	protected EntityManager em;
	
	public LineItem addLineItem(int customerId, int productId, long quantity)
			throws CustomerDoesNotExistException, OrderNotOpenException,
			ProductDoesNotExistException,GeneralPersistenceException, InvalidQuantityException {
		Product product = em.find(Product.class,productId);
		if(quantity <= 0 ) throw new InvalidQuantityException();
		if(product == null) throw new ProductDoesNotExistException();
		AbstractCustomer customer = loadCustomer(customerId);
		Order existingOpenOrder = customer.getOpenOrder();
		if(existingOpenOrder == null)
		{
			throw new OrderNotOpenException();
		}
		
		BigDecimal amount = product.getPrice().multiply(new BigDecimal(quantity));
		existingOpenOrder.setTotal(amount.add(existingOpenOrder.getTotal()));
		
		LineItemId lineItemId = new LineItemId();
		lineItemId.setProductId(productId);
		lineItemId.setOrderId(existingOpenOrder.getOrderId());
		LineItem existingLineItem = em.find(LineItem.class,lineItemId);
		if(existingLineItem == null)
		{
			LineItem lineItem = new LineItem();
			lineItem.setOrderId(existingOpenOrder.getOrderId());
			lineItem.setProductId(product.getProductId());
			lineItem.setAmount(amount);
			lineItem.setProduct(product);
			lineItem.setQuantity(quantity);
			em.persist(lineItem);
			return lineItem;
		}
		else
		{
			existingLineItem.setQuantity(existingLineItem.getQuantity() + quantity);
			existingLineItem.setAmount(existingLineItem.getAmount().add(amount));
			return existingLineItem;
		}
	}

	public Order openOrder(int customerId)
			throws CustomerDoesNotExistException, OrderAlreadyOpenException ,GeneralPersistenceException{
		AbstractCustomer customer = loadCustomer(customerId);
		Order existingOpenOrder = customer.getOpenOrder();
		if(existingOpenOrder != null)
		{
			throw new OrderAlreadyOpenException();
		}
		
		Order newOrder = new Order();
		newOrder.setCustomer(customer);
		newOrder.setStatus(Order.Status.OPEN);
		System.out.println(newOrder.getStatus());
		newOrder.setTotal(new BigDecimal(0));
		
		em.persist(newOrder);
		
		customer.setOpenOrder(newOrder);
		return newOrder;
	}

	public void submit(int customerId) throws CustomerDoesNotExistException,
			OrderNotOpenException, NoLineItemsException,GeneralPersistenceException {
		AbstractCustomer customer = loadCustomer(customerId);
		Order existingOpenOrder = customer.getOpenOrder();
		if(existingOpenOrder == null || existingOpenOrder.getStatus() != Order.Status.OPEN)
		{
			throw new OrderNotOpenException();
		}
		if(existingOpenOrder.getLineitems() == null || existingOpenOrder.getLineitems().size() <= 0 )
			throw new NoLineItemsException();
		
		existingOpenOrder.setStatus(Order.Status.SUBMITTED);
		customer.setOpenOrder(null);
	}

	public AbstractCustomer loadCustomer(int customerId) throws CustomerDoesNotExistException,GeneralPersistenceException {
		// TODO Auto-generated method stub
		AbstractCustomer customer = em.find(AbstractCustomer.class, customerId);
		 return customer;
	}

	public void removeLineItem(int customerId, int productId) throws CustomerDoesNotExistException, OrderNotOpenException, ProductDoesNotExistException, NoLineItemsException, GeneralPersistenceException {
		Product product = em.find(Product.class,productId);
		if(product == null) throw new ProductDoesNotExistException();
		
		AbstractCustomer customer = loadCustomer(customerId);
		Order existingOpenOrder = customer.getOpenOrder();
		if(existingOpenOrder == null || existingOpenOrder.getStatus() != Order.Status.OPEN)
		{
			throw new OrderNotOpenException();
		}
		//ORM Way
		
		LineItemId lineItemId = new LineItemId();
		lineItemId.setProductId(productId);
		lineItemId.setOrderId(existingOpenOrder.getOrderId());
		LineItem existingLineItem = em.find(LineItem.class,lineItemId);
		if(existingLineItem != null)
		{
			em.remove(existingLineItem);
		}
		else
		{
			throw new NoLineItemsException();
		}
		
		
		//Bulk Delete Feature, Better Performance
		/*
		Query query = em.createQuery("DELETE FROM LineItem l where l.productId = :productId and l.orderId = :orderId");
		query.setParameter("productId", productId);
		query.setParameter("orderId", existingOpenOrder.getOrderId());
		query.executeUpdate();*/
	}

}
