package org.pwte.example.hibernate.services;

import java.math.BigDecimal;
import java.util.HashSet;

import org.hibernate.HibernateException;
import org.hibernate.ObjectNotFoundException;
import org.hibernate.classic.Session;
import org.pwte.example.domain.AbstractCustomer;
import org.pwte.example.domain.LineItem;
import org.pwte.example.domain.LineItemId;
import org.pwte.example.domain.Order;
import org.pwte.example.domain.Product;
import org.pwte.example.domain.Order.Status;
import org.pwte.example.exception.CustomerDoesNotExistException;
import org.pwte.example.exception.GeneralPersistenceException;
import org.pwte.example.exception.NoLineItemsException;
import org.pwte.example.exception.OrderAlreadyOpenException;
import org.pwte.example.exception.OrderNotOpenException;
import org.pwte.example.exception.ProductDoesNotExistException;
import org.pwte.example.hibernate.HibernateUtil;
import org.pwte.example.service.CustomerOrderServices;

public class CustomerOrderServicesHibernateImpl implements CustomerOrderServices {

	public LineItem addLineItem(int customerId, int productId, long quantity)
			throws CustomerDoesNotExistException, OrderNotOpenException,
			ProductDoesNotExistException, GeneralPersistenceException {
		try {
			
			// 1.  Get a Hibernate Session
			Session session = HibernateUtil.getSessionFactory()
					.getCurrentSession();
			session.beginTransaction();

			// 2. Load the appropriate instance of the customer from the customerId 
			AbstractCustomer customer = null;
			try {
				customer = (AbstractCustomer) session.load(
					AbstractCustomer.class, new Integer(customerId));
			} catch (ObjectNotFoundException e) {
				throw new CustomerDoesNotExistException(
						"No customer exists matching id " + customerId);
			}
			
			// 3. Get the Customer's open order
			Order openOrder = customer.getOpenOrder();
			if (openOrder == null)
				throw new OrderNotOpenException(
						"Order not open for customer matching id "
								+ customerId);

			// 4. Get the Product from the product Id
			Product product = null;
			try {
				product = (Product) session.load(
					Product.class, new Integer(productId));
			} catch (ObjectNotFoundException e) {
				throw new ProductDoesNotExistException("No product exists matching id "+ productId);
			}

			// 5. Calculate the amount of the line item from the price and quantity
			BigDecimal amount = product.getPrice().multiply(new BigDecimal(quantity));
			
			// 6. Create and populate the new LineItemId

			LineItemId lineItemId = new LineItemId();
			lineItemId.setOrderId(openOrder.getOrderId());
			lineItemId.setProductId(product.getProductId());
			
			// 7. Check to see if the line item already exists in the session
			
			LineItem lineItem = null;
			try {
				lineItem = (LineItem) session.load(LineItem.class, lineItemId);
			} catch (ObjectNotFoundException e) {
				lineItem = new LineItem();
				lineItem.setLineItemId(lineItemId);
				lineItem.setProduct(product);
			}

			// 8. update the amount and quantity
			lineItem.setQuantity(quantity);
			lineItem.setAmount(amount);
			
			// 9. Set the list of line items if needed and add the new line Item to the list
			if (openOrder.getLineitems() == null) {
				openOrder.setLineitems(new HashSet<LineItem>());
			}
			openOrder.getLineitems().add(lineItem);
			
			// 10. Commit the transaction
			session.getTransaction().commit();

			return lineItem;
			
		} catch (HibernateException e) {
			throw new GeneralPersistenceException("Unexpected Exception " + e);
		}

	}

	public Order openOrder(int customerId)
			throws CustomerDoesNotExistException, OrderAlreadyOpenException,
			GeneralPersistenceException {

		try {
			// 1.  Get a Hibernate Session
			Session session = HibernateUtil.getSessionFactory()
					.getCurrentSession();
			session.beginTransaction();

			// 2. Load the appropriate instance of the customer from the customerId 
			AbstractCustomer customer = null;
			try {
				customer = (AbstractCustomer) session.load(
					AbstractCustomer.class, new Integer(customerId));
			} catch (ObjectNotFoundException e) {
				throw new CustomerDoesNotExistException(
						"No customer exists matching id " + customerId);
			}
			
			// 3. Check to see if the customer has an open order
			if (customer.getOpenOrder() != null)
				throw new OrderAlreadyOpenException(
						"Order already open for customer matching id "
								+ customerId);

			// 4. Create the new order and save it
			Order myOrder = new Order();
			myOrder.setStatus(Status.OPEN);
			myOrder.setCustomer(customer);
			myOrder.setTotal(new BigDecimal(0));
			session.save(myOrder);

			// 5. Update the open order in the customer
			customer.setOpenOrder(myOrder);
			
			// 6. Commit the transaction
			session.getTransaction().commit();

			return myOrder;
		} catch (HibernateException e) {
			throw new GeneralPersistenceException("Unexpected Exception " + e);
		}

	}

	public void submit(int customerId) throws CustomerDoesNotExistException,
			OrderNotOpenException, NoLineItemsException, GeneralPersistenceException {
		try {
			
			// 1. Get a Hibernate session
			Session session = HibernateUtil.getSessionFactory()
					.getCurrentSession();
			session.beginTransaction();

			// 2. Find the customer for this customerId
			AbstractCustomer customer = null;
			try {
				customer = (AbstractCustomer) session.load(
					AbstractCustomer.class, new Integer(customerId));
			} catch (ObjectNotFoundException e) {
				throw new CustomerDoesNotExistException(
						"No customer exists matching id " + customerId);
			}
			
			// 3. Get the customer's open order
			Order openOrder = customer.getOpenOrder();
			
			// 4. Throw an exception if the customer has no open order
			if (openOrder == null)
				throw new OrderNotOpenException(
						"Order not open for customer matching id "
								+ customerId);

			// 5. Throw an exception if the order has no line items
			if ((openOrder.getLineitems() == null) || openOrder.getLineitems().size() == 0)
				throw new NoLineItemsException("No line items for open order");
			
			// 6. Set the status to submitted
			openOrder.setStatus(Status.SUBMITTED);
			
			// 7. Set the open order in the customer to NULL and save the customer
			customer.setOpenOrder(null);
			session.save(customer);
			
			// 8. Commit the transaction
			session.getTransaction().commit();

		} catch (HibernateException e) {
			throw new GeneralPersistenceException("Unexpected Exception " + e);
		}

	}

	public AbstractCustomer loadCustomer(int customerId) throws CustomerDoesNotExistException, GeneralPersistenceException {
		try {
		
		// 1. Get a Hibernate session
		Session session = HibernateUtil.getSessionFactory()
				.getCurrentSession();
		session.beginTransaction();


		// 2. Find the customer for this customerId
		AbstractCustomer customer = null;
		try {
			customer = (AbstractCustomer) session.load(
				AbstractCustomer.class, new Integer(customerId));
		} catch (ObjectNotFoundException e) {
			throw new CustomerDoesNotExistException(
					"No customer exists matching id " + customerId);
		}
		
		// 3. Commit the transaction
		session.getTransaction().commit();
		
		return customer;
		
		} catch (HibernateException e) {
			throw new GeneralPersistenceException("Unexpected Exception " + e);
		}
	}

	@Override
	public void removeLineItem(int customerId, int productId)
			throws CustomerDoesNotExistException, OrderNotOpenException,
			ProductDoesNotExistException, NoLineItemsException,
			GeneralPersistenceException {
		// TODO Auto-generated method stub
		// 1. Get a Hibernate session
		try
		{
		Session session = HibernateUtil.getSessionFactory()
				.getCurrentSession();
		session.beginTransaction();
		
		// 2. Load the appropriate instance of the customer from the customerId 
		AbstractCustomer customer = null;
		try {
			customer = (AbstractCustomer) session.load(
				AbstractCustomer.class, new Integer(customerId));
		} catch (ObjectNotFoundException e) {
			throw new CustomerDoesNotExistException(
					"No customer exists matching id " + customerId);
		}
		
		// 3. Get the Customer's open order
		Order openOrder = customer.getOpenOrder();
		if (openOrder == null)
			throw new OrderNotOpenException(
					"Order not open for customer matching id "
							+ customerId);

		// 4. Get the Product from the product Id
		Product product = null;
		try {
			product = (Product) session.load(
				Product.class, new Integer(productId));
		} catch (ObjectNotFoundException e) {
			throw new ProductDoesNotExistException("No product exists matching id "+ productId);
		}

		// 5. Create and populate the new LineItemId

		LineItemId lineItemId = new LineItemId();
		lineItemId.setOrderId(openOrder.getOrderId());
		lineItemId.setProductId(product.getProductId());
		
		// 6. Check to see if the line item already exists in the session
		
		LineItem lineItem = null;
		try {
			lineItem = (LineItem) session.load(LineItem.class, lineItemId);
		} catch (ObjectNotFoundException e) {
			throw new NoLineItemsException(e);
		}
		
		// 7. remove from Order
		openOrder.getLineitems().remove(lineItem);
				
		// 8. Commit the transaction
		session.getTransaction().commit();
		} catch (HibernateException e) {
			e.printStackTrace(System.out);
			throw new GeneralPersistenceException("Unexpected Exception " + e);
		}
	}

}
