package org.pwte.example.ibatis;

import java.io.IOException;
import java.io.Reader;
import java.math.BigDecimal;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.pwte.example.domain.AbstractCustomer;
import org.pwte.example.domain.LineItem;
import org.pwte.example.domain.Order;
import org.pwte.example.domain.Product;
import org.pwte.example.exception.CustomerDoesNotExistException;
import org.pwte.example.exception.GeneralPersistenceException;
import org.pwte.example.exception.NoLineItemsException;
import org.pwte.example.exception.OrderAlreadyOpenException;
import org.pwte.example.exception.OrderNotOpenException;
import org.pwte.example.exception.ProductDoesNotExistException;
import org.pwte.example.service.CustomerOrderServices;

import com.ibatis.common.resources.Resources;
import com.ibatis.sqlmap.client.SqlMapClient;
import com.ibatis.sqlmap.client.SqlMapClientBuilder;

public class CustomerOrderServicesiBatisImpl implements CustomerOrderServices {

	private SqlMapClient dataMapper = null;

	public AbstractCustomer loadCustomer(int customerId) throws CustomerDoesNotExistException, GeneralPersistenceException {
		
		// 1. (Re-)Initialize iBatis Data Mapper
		SqlMapClient dm = getDataMapper();

		// 2. Retrieve Customer
		AbstractCustomer customer = null;
		try {
			// N+1 selects 
			/*
			customer = (AbstractCustomer) dm.queryForObject("customer.select", customerId);
			*/
			
			//Single Query
			List<AbstractCustomer> list = dm.queryForList("customer.with.openOrder.select", customerId);
			if(list.size() > 0)
			customer = list.get(0);
		
		} catch(SQLException sqle) {
			throw new GeneralPersistenceException(sqle);
		}
		
		return customer;
	}
	
	public Order openOrder(int customerId) throws CustomerDoesNotExistException,
	OrderAlreadyOpenException, GeneralPersistenceException {
			
		// 1. (Re-)Initialize iBatis Data Mapper
		SqlMapClient dm = getDataMapper();
		
		Order order = null;
		try {
			dm.startTransaction();
		
			// 2. Retrieve Customer
			AbstractCustomer customer = loadCustomer(customerId);
				
			// 3. Ensure Customer exists
			if(customer == null) throw new CustomerDoesNotExistException();
			
			System.out.println("OPEN ORDER " + customer.getOpenOrder().getOrderId());
			// 4. Ensure Customer does not already have Order in 'OPEN' state
			if(customer.getOpenOrder() != null && customer.getOpenOrder().getOrderId() != 0) throw new OrderAlreadyOpenException();
			
			// 5. Create and persist Order
			order = new Order();
			order.setCustomer(customer);
			order.setStatus(Order.Status.OPEN);
			order.setTotal(new BigDecimal(0.00));
			
			int orderId = (Integer) dm.insert("order.open", order);
			order.setOrderId(orderId);
				
			// 6. Update Customer with reference to 'OPEN' Order
			dm.update("customer.open.order", customerId);
		
			dm.commitTransaction();
			
		} catch(SQLException sqle) {
			throw new GeneralPersistenceException(sqle);
		} finally {
			try {
				dm.endTransaction();
			} catch(SQLException sqle) {
				throw new GeneralPersistenceException(sqle);
			}
		}
			
		return order;
	}
		
	public LineItem addLineItem(int customerId, int productId, long quantity)
			throws CustomerDoesNotExistException, OrderNotOpenException, ProductDoesNotExistException, GeneralPersistenceException {
		
		// 1. (Re-)Initialize iBatis Data Mapper
		SqlMapClient dm = getDataMapper();
		
		LineItem lineitem = null;
		try {
			dm.startTransaction();
			
			// 2. Retrieve Product
			Product product = null;
			product = (Product) dm.queryForObject("product.select", productId);
			
			// 3. Ensure Product exists
			if(product == null) throw new ProductDoesNotExistException();
			
			// 4. Retrieve Customer
			AbstractCustomer customer = loadCustomer(customerId);
				
			// 5. Ensure Customer exists
			if(customer == null) throw new CustomerDoesNotExistException();
			
			// 6. Ensure Customer has an 'OPEN' Order
			if (customer.getOpenOrder() == null || customer.getOpenOrder().getOrderId() == 0) throw new OrderNotOpenException();
			
			// 7. Check whether Order has existing LineItem
			Map keys = new HashMap();
			keys.put("orderId", customer.getOpenOrder().getOrderId());
			keys.put("productId", productId);
			lineitem = null;
			lineitem = (LineItem) dm.queryForObject("lineitem.select", keys);
			
			// 8. If LineItem already exists update existing quantity and amount
			if(lineitem != null) {
				lineitem.setQuantity(lineitem.getQuantity() + quantity);
				lineitem.setAmount(product.getPrice().multiply(new BigDecimal(lineitem.getQuantity())));
				Map wrapper = new HashMap();
				wrapper.put("lineitem", lineitem);
				wrapper.put("orderId", customer.getOpenOrder().getOrderId());
				// 9. Update LineItem
				dm.update("lineitem.update", wrapper);
				
			// 10. If LineItem does not exist, create it
			} else {
				BigDecimal amount = product.getPrice().multiply(new BigDecimal(quantity));
				
				lineitem = new LineItem();
				lineitem.setProduct(product);
				lineitem.setQuantity(quantity);
				lineitem.setAmount(amount);
				
				// 11. Persist LineItem
				Map wrapper = new HashMap();
				wrapper.put("orderId", customer.getOpenOrder().getOrderId());
				wrapper.put("lineitem", lineitem);
				dm.insert("lineitem.insert", wrapper);
			}
			
			dm.commitTransaction();
		
		}catch(SQLException sqle) {
			throw new GeneralPersistenceException(sqle);
		} finally{
			try {
				dm.endTransaction();
			} catch(SQLException sqle) {
				throw new GeneralPersistenceException(sqle);
			}
		}
			
		return lineitem;
	}
	
	public void submit(int customerId) throws CustomerDoesNotExistException,
			OrderNotOpenException, NoLineItemsException, GeneralPersistenceException {
		
		// 1. (Re-)Initialize iBatis Data Mapper
		SqlMapClient dm = getDataMapper();

		try {
			dm.startTransaction();
		
			// 2. Retrieve Customer
			AbstractCustomer customer = loadCustomer(customerId);
				
			// 3. Ensure Customer exists
			if(customer == null) throw new CustomerDoesNotExistException();
			
			// 4. Ensure Customer has 'OPEN' Order
			if(customer.getOpenOrder() == null || customer.getOpenOrder().getOrderId() == 0) throw new OrderNotOpenException();
			
			System.out.println(customer.getOpenOrder().getLineitems().size());
			// 5. Ensure Order has LineItems
			if(customer.getOpenOrder().getLineitems() == null || customer.getOpenOrder().getLineitems().size() == 0 || (((LineItem)(customer.getOpenOrder().getLineitems().get(0))).getProduct()) == null) throw new NoLineItemsException();
			
			// 6. Update Order status to 'SUBMITTED'
			dm.update("order.submit", customer.getOpenOrder().getOrderId());
							
			// 7. Remove Customer openOrder reference
			dm.update("customer.close.order", customerId);
		
			dm.commitTransaction();
			
		}catch(SQLException sqle) {
			throw new GeneralPersistenceException(sqle);
		} finally{
			try {
				dm.endTransaction();
			} catch(SQLException sqle) {
				throw new GeneralPersistenceException(sqle);
			}
		}
	}
	
	public void removeLineItem(int customerId, int productId) throws CustomerDoesNotExistException, OrderNotOpenException, ProductDoesNotExistException, NoLineItemsException, GeneralPersistenceException {
	
		// 1. (Re-)Initialize iBatis Data Mapper
		SqlMapClient dm = getDataMapper();
		
		try {
			dm.startTransaction();
		
			// 2. Retrieve Customer
			AbstractCustomer customer = loadCustomer(customerId);
				
			// 3. Ensure Customer exists
			if(customer == null) throw new CustomerDoesNotExistException();
			
			// 4. Ensure Customer has 'OPEN' Order
			if(customer.getOpenOrder() == null || customer.getOpenOrder().getOrderId() == 0) throw new OrderNotOpenException();
			
			// 5. Ensure Order has LineItems
			if(customer.getOpenOrder().getLineitems() == null) throw new NoLineItemsException();
			
			// 6. Lookup Product
			Product product;
			product = (Product) dm.queryForObject("product.select", productId);
			
			// 7. Ensure Product Exists
			if(product == null) throw new ProductDoesNotExistException();
				
			// 8. Create Key Class for Deletion
			HashMap keys = new HashMap();
			keys.put("orderId", customer.getOpenOrder().getOrderId());
			keys.put("productId", productId);
			
			// 9. Delete LineItem
			dm.delete("lineitem.delete", keys);
			
			dm.commitTransaction();
			
		}catch(SQLException sqle) {
			throw new GeneralPersistenceException(sqle);
		} finally{
			try {
				dm.endTransaction();
			} catch(SQLException sqle) {
				throw new GeneralPersistenceException(sqle);
			}
		}
	}
	
	/**
	 * utility method that retrieves iBatis data mapper interface
	 * @return SqlMapClient
	 * @throws GeneralPersistenceException
	 */
	private SqlMapClient getDataMapper() throws GeneralPersistenceException {
		if(dataMapper == null) {
			Reader reader = null;
			try {
				reader = Resources.getResourceAsReader("sql-map-config.xml");
			} catch (IOException ioe) {
				throw new GeneralPersistenceException("Database Connection Error: " + ioe.getMessage(), ioe);
			}
			dataMapper = SqlMapClientBuilder.buildSqlMapClient(reader);
		}
	    return dataMapper;
	}
}
