package org.pwte.example.jdbc;

import java.math.BigDecimal;
import java.sql.*;
import java.util.*;

import org.pwte.example.domain.*;
import org.pwte.example.exception.*;
import org.pwte.example.jdbc.util.JDBCHelper;
import org.pwte.example.service.CustomerOrderServices;

public class CustomerOrderServicesJDBCImpl implements CustomerOrderServices {
	 // create a new residential customer
    public void createResidentialCustomer(ResidentialCustomer customer) throws GeneralPersistenceException
    {
    	PreparedStatement insertCustomerPS = null;
    	Connection conn = null;
    	try 
    	{
    		// 1. Setup connection
    		conn = JDBCHelper.getJDBCConnection();
    		// 2. Populate insert query
			String insertCustomerSQL = "insert into CUSTOMER ( NAME, RESIDENTIAL_HOUSEHOLD_SIZE," +
				"RESIDENTIAL_FREQUENT_CUSTOMER, CUSTOMER_ID, TYPE ) values ( ?, ?, ?, ?, ? )";
			insertCustomerPS = conn.prepareStatement(insertCustomerSQL);
			insertCustomerPS.setString(1, customer.getName());
			insertCustomerPS.setInt(2, customer.getHouseholdSize());
			boolean frequentCustomerEval = customer.isFrequentCustomer();
			if (frequentCustomerEval) insertCustomerPS.setString(3, "Y");
			else if (!frequentCustomerEval) insertCustomerPS.setString(3, "N");	
			insertCustomerPS.setInt(4, customer.getCustomerId());
			insertCustomerPS.setString(5,"RESIDENTIAL");
			// 3. Issue update statement
			insertCustomerPS.executeUpdate();
		} 
    	catch (SQLException sqle) 
    	{
			try {
				if (conn!=null) conn.rollback();
			} catch (SQLException sqlex) 
			{
				sqlex.printStackTrace();
				throw new GeneralPersistenceException();
			}
		} 
    	finally
		{
    		JDBCHelper.closeStatement(insertCustomerPS);
    		JDBCHelper.closeConnection(conn);
		}
    }
    
    // create a new order
	public Order openOrder(int customerId) 
		throws CustomerDoesNotExistException, OrderAlreadyOpenException, GeneralPersistenceException 
	{
		Connection conn = null;
		ResultSet retrievedOrdersResultSet = null;
		PreparedStatement insertOrderPS = null;
		PreparedStatement updateCustomerPS = null;
		PreparedStatement retrieveOrdersIdPS = null;
		
		Order order = null;
		try
		{
			// 1. Setup connection
			conn = JDBCHelper.getJDBCConnection();
		
			// 2. Load customer using customer id
			AbstractCustomer customer = loadCustomer(customerId);
			 			
			// 3. Get assigned open order of customer
			Order openOrder = customer.getOpenOrder();
			
			// 4. Ensure order is not already open
			if (openOrder!=null) throw new OrderAlreadyOpenException();
					
			// 5. Create Order
			order = new Order();
			order.setCustomer(customer);
			order.setStatus(Order.Status.OPEN);
			order.setTotal(new BigDecimal(0.00));
						
			// 6. Persist Order - assuming key generation by DB for ORDER_ID
			insertOrderPS = conn.prepareStatement("insert into ORDERS (CUSTOMER_ID,STATUS,TOTAL) values (?,?,?)");
			insertOrderPS.setInt(1, customerId); // set customerID
			insertOrderPS.setString(2,"OPEN"); // set status
			insertOrderPS.setBigDecimal(3,new BigDecimal(0.00));  // set total
			insertOrderPS.executeUpdate(); // execute update statement
			insertOrderPS.close();
			
			//  7.Fetch the order Id
			retrieveOrdersIdPS = 
				conn.prepareStatement("select ORDER_ID from ORDERS where CUSTOMER_ID=? and STATUS = 'OPEN'");
			retrieveOrdersIdPS.setInt(1, customerId);
			ResultSet rs = retrieveOrdersIdPS.executeQuery();
			
			if(rs.next())
			{
				int orderId = rs.getInt("ORDER_ID");
				order.setOrderId(orderId);
			}
			else
			{
				throw new GeneralPersistenceException("Could not retrieve new order id");
			}
			
			//	8. Update Customer with newly assigned order ID
			updateCustomerPS 
				= conn.prepareStatement("update CUSTOMER set OPEN_ORDER=? where CUSTOMER_ID=?");
			updateCustomerPS.setInt(1, order.getOrderId());
			updateCustomerPS.setInt(2, customerId);
			updateCustomerPS.executeUpdate();
			updateCustomerPS.close();
			
		    conn.commit();
		}
		catch (SQLException sqle)
		{
			sqle.printStackTrace();
			try 
			{
				if (conn!=null) conn.rollback();
			} 
			catch (SQLException e) 
			{
				e.printStackTrace();
			}
			throw new GeneralPersistenceException();
		}
		finally
		{
			JDBCHelper.closeResultSet(retrievedOrdersResultSet);
			JDBCHelper.closeStatement(retrieveOrdersIdPS);
			JDBCHelper.closeStatement(updateCustomerPS);
			JDBCHelper.closeStatement(retrieveOrdersIdPS);
		}
		return order;
	}
    

	public LineItem addLineItem(int customerId, int productId, long quantity)
			throws CustomerDoesNotExistException, OrderNotOpenException,
			ProductDoesNotExistException, GeneralPersistenceException 
	{
		Connection conn = null;
		ResultSet productResultSet = null;
		ResultSet customerResultSet = null;
		ResultSet lineItemResultSet = null;
		PreparedStatement productPS = null;
		PreparedStatement customerPS = null;
		PreparedStatement lineItemPS = null;
		PreparedStatement lineItemInsertPS = null;
		PreparedStatement lineItemUpdatePS = null;
		
		LineItem item = new LineItem();
		try {
			
			// 1. Setup connection
			conn = JDBCHelper.getJDBCConnection();
			
			// 2. Retrieve Product
			productPS 
				= conn.prepareStatement("select PRICE,DESCRIPTION from PRODUCT where PRODUCT_ID=?");
			productPS.setInt(1, productId);
			productResultSet = productPS.executeQuery();
			
			// 3. Ensure product exists
			if (productResultSet==null) throw new ProductDoesNotExistException();
			
			// 4. Retrieve Customer
			customerPS = conn.prepareStatement("select OPEN_ORDER from CUSTOMER where CUSTOMER_ID=?");
			customerPS.setInt(1, customerId);
			customerResultSet = customerPS.executeQuery();
			
			// 5. Ensure Customer Exists
			if (customerResultSet == null) throw new CustomerDoesNotExistException();
			customerResultSet.next();
			
			// 6. Ensure Customer has an 'OPEN' Order
			Object openOrder = customerResultSet.getObject("OPEN_ORDER");
			
			if (customerResultSet.getObject("OPEN_ORDER")==null) throw new OrderNotOpenException();
			int orderID = (Integer)openOrder;
						
			// 7. calculate cost for product addition, but first make sure product exists
			if (!productResultSet.next()) 
			{
				productResultSet.close();
				productPS.close();
				customerResultSet.close();
				customerPS.close();
				conn.commit();
				conn.close();
				throw new ProductDoesNotExistException();
			}
			BigDecimal productPrice = productResultSet.getBigDecimal("PRICE");
			BigDecimal additionalCost = productPrice.multiply(new BigDecimal(quantity));
			BigDecimal lineItemAmount = new BigDecimal(0.0);
						
			// 8. Determine if any existing line items exist for the current product  
			lineItemPS 
			    = conn.prepareStatement("select QUANTITY, AMOUNT from LINE_ITEM where ORDER_ID = ? and PRODUCT_ID = ?");
			lineItemPS.setInt(1, orderID);
			lineItemPS.setInt(2, productId);
			lineItemResultSet = lineItemPS.executeQuery();
			
			boolean lineItemsCheck = lineItemResultSet.next();
			if (!lineItemsCheck) // check if any line items already exist for product
			{
				// line item does not exist for product yet
				// Calculate amount = price * quantity
				lineItemAmount = additionalCost;
				// create a new line item
				lineItemInsertPS 
				  = conn.prepareStatement("insert into LINE_ITEM(ORDER_ID,PRODUCT_ID,QUANTITY,AMOUNT) values(?,?,?,?)");
				int orderId = customerResultSet.getInt("OPEN_ORDER");
				lineItemInsertPS.setInt(1,orderId);
				lineItemInsertPS.setInt(2,productId);
				lineItemInsertPS.setLong(3,quantity);
				lineItemInsertPS.setBigDecimal(4,lineItemAmount);
				lineItemInsertPS.executeUpdate();
				lineItemInsertPS.close();

			}
			else 
			{
				// line item already exists
				int existingQuantity = lineItemResultSet.getInt("QUANTITY");
				BigDecimal existingAmount = lineItemResultSet.getBigDecimal("AMOUNT");
				quantity+=existingQuantity;
				lineItemAmount = existingAmount.add(additionalCost);
				// update existing line item
				lineItemUpdatePS 
				  = conn.prepareStatement("update LINE_ITEM set QUANTITY = ?, AMOUNT = ? where ORDER_ID = ? and PRODUCT_ID = ?");
				int orderId = customerResultSet.getInt("OPEN_ORDER");
				lineItemUpdatePS.setLong(1,quantity);
				lineItemUpdatePS.setBigDecimal(2,lineItemAmount);
				lineItemUpdatePS.setLong(3,orderId);
				lineItemUpdatePS.setInt(4,productId);
				lineItemUpdatePS.executeUpdate();
				lineItemUpdatePS.close();
			}
			
			item.setQuantity(quantity);
			item.setAmount(lineItemAmount);
			
			Product product = new Product();
			product.setProductId(productId); 
			product.setDescription(productResultSet.getString("DESCRIPTION"));
			product.setPrice(productResultSet.getBigDecimal("PRICE"));
			item.setProduct(product); // populate product into item
			
			// 9. Perform clean up
			conn.commit();			
		} 
		catch (SQLException e) 
		{
			e.printStackTrace();
			try 
			{
				if (conn!=null) conn.rollback();
			} 
			catch (SQLException sqle) 
			{
				sqle.printStackTrace();
			}
			throw new GeneralPersistenceException();
		}
		finally
		{
			JDBCHelper.closeResultSet(customerResultSet);
			JDBCHelper.closeResultSet(lineItemResultSet);
			JDBCHelper.closeResultSet(productResultSet);
			JDBCHelper.closeStatement(lineItemPS);
			JDBCHelper.closeStatement(customerPS);
			JDBCHelper.closeStatement(productPS);
			JDBCHelper.closeConnection(conn);
		}
		return item;
		
	}
	

	public void submit(int customerId) throws CustomerDoesNotExistException,
			OrderNotOpenException, NoLineItemsException, GeneralPersistenceException 
	{
		Connection conn = null;
		PreparedStatement lineItemPS = null;
		PreparedStatement updateOrderStatusPS = null;
		PreparedStatement customerCloseOrderPS = null;
			
		try {
			// 1. Setup connection
			conn = JDBCHelper.getJDBCConnection();
			
			// 2. Retrieve customer using customer ID
			AbstractCustomer customer = loadCustomer(customerId);
			
			// 3. Ensure customer has an 'OPEN' order, else throw exception 
			Order openOrder = customer.getOpenOrder();
			if (openOrder==null) throw new OrderNotOpenException();
			

			// 4. Ensure Order has LineItems, if not throw exception
			if ((openOrder.getLineitems()==null) || (openOrder.getLineitems().size()<=0)) throw new NoLineItemsException();
			
			// 5. Update Order Status to Submitted
			updateOrderStatusPS 
				= conn.prepareStatement("update ORDERS set STATUS ='SUBMITTED' WHERE ORDER_ID=?");
			int orderId = openOrder.getOrderId();
			updateOrderStatusPS.setInt(1, orderId);
			updateOrderStatusPS.executeUpdate();
			updateOrderStatusPS.close();
			
			// 6. Remove Customer openOrder reference
			customerCloseOrderPS
				= conn.prepareStatement("update CUSTOMER set OPEN_ORDER = NULL where CUSTOMER_ID=?");
			customerCloseOrderPS.setInt(1, customerId);
			customerCloseOrderPS.executeUpdate();
			customerCloseOrderPS.close();
			conn.commit();			
		} 
		catch (SQLException e) 
		{
			e.printStackTrace();
			try 
			{
				if (conn!=null) conn.rollback();
			} catch (SQLException sqle) 
			{
				sqle.printStackTrace();
			}
			throw new GeneralPersistenceException();
		} 
		finally 
		{
			JDBCHelper.closeStatement(lineItemPS);
			JDBCHelper.closeStatement(updateOrderStatusPS);
			JDBCHelper.closeStatement(customerCloseOrderPS);
			JDBCHelper.closeConnection(conn);
		}
	}
	
	public void setBusinessPartnerToFalse(int customerID) throws GeneralPersistenceException
	{
		Connection conn = null;
		PreparedStatement updateBusinessPartnerPS = null;
		try 
		{
			// 1. Set up connection
			conn = JDBCHelper.getJDBCConnection();
			
			// 2. Create update statement
			updateBusinessPartnerPS = conn.prepareStatement("UPDATE CUSTOMER SET BUSINESS_PARTNER='N' WHERE CUSTOMER_ID=?");
			
			// 3. Populate statement
			updateBusinessPartnerPS.setInt(1, customerID);
			
			// 4. Issue update statement
			updateBusinessPartnerPS.executeUpdate();
			conn.commit();
		} 
		catch (SQLException e) 
		{
			try 
			{
				e.printStackTrace();
				if (conn!=null) conn.rollback();
			} 
			catch (SQLException sqle) 
			{
				sqle.printStackTrace();
				throw new GeneralPersistenceException();
			}
			
		}
		finally
		{
			JDBCHelper.closeStatement(updateBusinessPartnerPS);
			JDBCHelper.closeConnection(conn);
		}
		
		
	}

	public AbstractCustomer loadCustomer(int customerId) throws CustomerDoesNotExistException, GeneralPersistenceException 
	{
		PreparedStatement customerPS = null;
		ResultSet customerResultSet = null;
		AbstractCustomer customer = null;
		Connection conn = null;
		
		try {
			
			// 1. Setup connection
			conn = JDBCHelper.getJDBCConnection();
	    					
			// 2. Retrieve Customer
			customerPS 
    			= conn.prepareStatement("SELECT c.CUSTOMER_ID,c.OPEN_ORDER,c.NAME,c.BUSINESS_VOLUME_DISCOUNT,c.BUSINESS_PARTNER,c.BUSINESS_DESCRIPTION,c.RESIDENTIAL_HOUSEHOLD_SIZE,c.RESIDENTIAL_FREQUENT_CUSTOMER,c.TYPE,o.ORDER_ID,o.STATUS,o.TOTAL,l.PRODUCT_ID,l.ORDER_ID,l.QUANTITY,l.AMOUNT,p.PRODUCT_ID,p.PRICE,p.DESCRIPTION FROM CUSTOMER c LEFT OUTER JOIN ORDERS o ON c.OPEN_ORDER = o.ORDER_ID " +
    									"LEFT OUTER JOIN (LINE_ITEM l JOIN PRODUCT p ON l.PRODUCT_ID = p.PRODUCT_ID)ON o.ORDER_ID = l.ORDER_ID WHERE c.CUSTOMER_ID = ?");
			customerPS.setInt(1, customerId);
			customerResultSet = customerPS.executeQuery();

			// 3. Check if the query returned any results, if not throw exception
			if (customerResultSet.next()==false) 
			{
				conn.commit();
				conn.close();
				throw new CustomerDoesNotExistException();
			}
			
			
			String customerType = customerResultSet.getString("TYPE");
			
			// handle business customer type
			if (customerType.equals("BUSINESS"))
			{
				BusinessCustomer businessCustomer = new BusinessCustomer();
				String volDiscountEval = customerResultSet.getString("BUSINESS_VOLUME_DISCOUNT");
				if (volDiscountEval.equals("Y")) businessCustomer.setVolumeDiscount(true);
				else if (volDiscountEval.equals("N")) businessCustomer.setVolumeDiscount(false);
				String businessPartnerEval = customerResultSet.getString("BUSINESS_PARTNER");
				if (businessPartnerEval.equals("Y")) businessCustomer.setBusinessPartner(true);
				else if (businessPartnerEval.equals("N")) businessCustomer.setBusinessPartner(false);	
				customer = businessCustomer;
			}
			// handle residential customer type
			else if (customerType.equals("RESIDENTIAL"))
			{
				ResidentialCustomer residentialCustomer = new ResidentialCustomer();
				residentialCustomer.setHouseholdSize(customerResultSet.getShort("RESIDENTIAL_HOUSEHOLD_SIZE"));
				String frequentCustomerEval = customerResultSet.getString("RESIDENTIAL_FREQUENT_CUSTOMER");
				if (frequentCustomerEval.equals("Y")) residentialCustomer.setFrequentCustomer(true);
				else if (frequentCustomerEval.equals("N")) residentialCustomer.setFrequentCustomer(false);	
				customer = residentialCustomer;
				
			}
			// populate common customer fields
			customer.setCustomerId(customerId);
			customer.setName(customerResultSet.getString("NAME"));
			
			Order order = null;
			Object openOrder = customerResultSet.getObject("OPEN_ORDER");
			if (openOrder!=null) 
			{
				int openOrderID = (Integer)openOrder;
				order = new Order();
				order.setOrderId(openOrderID);
				order.setStatus(Order.Status.OPEN);
				customer.setOpenOrder(order);
			    order.setTotal(customerResultSet.getBigDecimal("TOTAL"));
			    order.setCustomer(customer);	
			    customer.setOpenOrder(order); 
				
			    HashSet<LineItem> orderLineItems = null;
			    Object productObject =  customerResultSet.getObject("PRODUCT_ID");
			    orderLineItems = new HashSet<LineItem>();
			    // populate order line items
			    if ((productObject!=null) && (order!=null))
			    {
			    	
			    	
			    	do 
			    	{
			    		LineItem lineItem = new LineItem();
			    		lineItem.setQuantity(customerResultSet.getLong("QUANTITY"));
			    		lineItem.setAmount(customerResultSet.getBigDecimal("AMOUNT"));
			    		Product product = new Product();
			    		product.setDescription(customerResultSet.getString("DESCRIPTION"));
			    		product.setPrice(customerResultSet.getBigDecimal("PRICE"));
			    		product.setProductId(customerResultSet.getInt("PRODUCT_ID"));
			    		lineItem.setProduct(product);
			    		orderLineItems.add(lineItem);
			    	}    
			    	while (customerResultSet.next());
						
			    }	
				order.setLineitems(orderLineItems);
				
			}
			
			customerResultSet.close();
			customerPS.close();
			conn.commit();
			conn.close();
			
		} 
		catch (SQLException e) 
		{
			e.printStackTrace();
			try 
			{
				if (conn!=null) conn.rollback();
			} 
			catch (SQLException sqle) 
			{
				sqle.printStackTrace();
			}
			throw new GeneralPersistenceException();
		} 
		finally
		{
			JDBCHelper.closeResultSet(customerResultSet);
			JDBCHelper.closeStatement(customerPS);
			JDBCHelper.closeConnection(conn);
		}
		return customer;
	}
	

	public void removeLineItem(int customerId, int productId ) throws CustomerDoesNotExistException, OrderNotOpenException, ProductDoesNotExistException, NoLineItemsException, GeneralPersistenceException
	{
		Connection conn = null;
		PreparedStatement productPS = null;
		PreparedStatement productDeletePS = null;
		ResultSet productResultSet = null;
		try 
		{
			// 1. Set up connection			
			conn = JDBCHelper.getJDBCConnection();
			
			// 2. Check if product exists
			productPS = conn.prepareStatement("SELECT * FROM PRODUCT WHERE PRODUCT_ID = ?");
			productPS.setInt(1,productId);
			productResultSet = productPS.executeQuery();
			
			boolean productExists = false;
			productExists = productResultSet.next();
			productResultSet.close();
			productPS.close();
			conn.commit();
			if (!productExists) throw new ProductDoesNotExistException();
			
			
			// 3. Load customer
			AbstractCustomer customer = loadCustomer(customerId);
			Order openOrder = customer.getOpenOrder();
			
			// 4. Check if open order exists, if not throw exception
			if (openOrder == null) 
			{
				throw new OrderNotOpenException();
			}
			int orderId = openOrder.getOrderId(); 
			
			// 5. Remove line item
			productDeletePS = conn.prepareStatement("DELETE FROM LINE_ITEM WHERE PRODUCT_ID = ? AND ORDER_ID = ?");
			productDeletePS.setInt(1,productId);
			productDeletePS.setInt(2,orderId);
			productDeletePS.executeUpdate();
			productDeletePS.close();
			conn.commit();
			conn.close();
			
		}
		catch (SQLException e) 
		{
			e.printStackTrace();
			try 
			{
			  if (conn!=null) conn.rollback();
			} 
			catch (SQLException sqle) 
			{
				sqle.printStackTrace();
				throw new GeneralPersistenceException();
			}
		}
		
		finally
		{
			JDBCHelper.closeResultSet(productResultSet);
			JDBCHelper.closeStatement(productPS);
			JDBCHelper.closeStatement(productDeletePS);
			JDBCHelper.closeConnection(conn);
		}
		
	}
	
	public void updateExistingLineItem(int orderID, int productID, int quantity) throws GeneralPersistenceException
	{
		Connection conn = null;
		PreparedStatement currentPricePS = null;
		PreparedStatement updateLineItemPS = null;
		
		try 
		{
			// 1. Set up connection
			conn = JDBCHelper.getJDBCConnection();
	
			// 2. Get the price of the product represented by the line item
			currentPricePS = conn.prepareStatement("SELECT PRICE FROM PRODUCT WHERE PRODUCT_ID = ?");
			currentPricePS.setInt(1,productID);
			ResultSet priceRS = currentPricePS.executeQuery();
			priceRS.next();
			BigDecimal productPrice = priceRS.getBigDecimal("PRICE");
				
			// 3. Calculate the new amount for the line item
				
			BigDecimal updateAmount = productPrice.multiply(BigDecimal.valueOf(quantity));
				
			// 4. Persist the new quantity and amount
			updateLineItemPS = conn.prepareStatement("UPDATE LINE_ITEM SET QUANTITY = ?, AMOUNT = ? WHERE ORDER_ID = ? AND PRODUCT_ID = ?");
			updateLineItemPS.setInt(1, quantity);
			updateLineItemPS.setBigDecimal(2, updateAmount);
			updateLineItemPS.setInt(3, orderID);
			updateLineItemPS.setInt(4, productID);
			updateLineItemPS.executeUpdate();
			conn.commit();
		}
		catch (SQLException e) 
		{
			try 
			{
				if (conn!=null) conn.rollback();
			} 
			catch (SQLException sqle) 
			{
				sqle.printStackTrace();
				throw new GeneralPersistenceException();
			}
		}
		finally
		{
			JDBCHelper.closeStatement(updateLineItemPS);
			JDBCHelper.closeStatement(currentPricePS);
			JDBCHelper.closeConnection(conn);
		}
	}
}
