package org.pwte.example.service.test;


import java.io.FileInputStream;
import java.util.Set;
import org.dbunit.DBTestCase;
import org.dbunit.PropertiesBasedJdbcDatabaseTester;
import org.dbunit.dataset.IDataSet;
import org.dbunit.dataset.xml.FlatXmlDataSet;
import org.dbunit.operation.DatabaseOperation;
import org.pwte.example.domain.AbstractCustomer;
import org.pwte.example.domain.BusinessCustomer;
import org.pwte.example.domain.LineItem;
import org.pwte.example.domain.Order;
import org.pwte.example.domain.ResidentialCustomer;
import org.pwte.example.exception.CustomerDoesNotExistException;
import org.pwte.example.exception.GeneralPersistenceException;
import org.pwte.example.exception.InvalidQuantityException;
import org.pwte.example.exception.NoLineItemsException;
import org.pwte.example.exception.OrderAlreadyOpenException;
import org.pwte.example.exception.OrderNotOpenException;
import org.pwte.example.exception.ProductDoesNotExistException;
import org.pwte.example.service.CustomerOrderServices;
import org.pwte.example.service.impl.CustomerOrderServiceImpl;


public class CustomerOrderServicesTest extends DBTestCase{

	private CustomerOrderServices customerOrderServices;
	private int customerId = 2;
	private int businessCustomerId = 3;
	
	public CustomerOrderServicesTest(String name)
	{
		super(name); 
		System.setProperty( PropertiesBasedJdbcDatabaseTester.DBUNIT_DRIVER_CLASS, "org.apache.derby.jdbc.ClientDriver" );
	    System.setProperty( PropertiesBasedJdbcDatabaseTester.DBUNIT_CONNECTION_URL, "jdbc:derby://localhost:1527/PWTE" );
	    System.setProperty( PropertiesBasedJdbcDatabaseTester.DBUNIT_SCHEMA, "APP" );
	}
	
	
	
	public void setUp() throws Exception {
		
		super.setUp();
		customerOrderServices = new CustomerOrderServiceImpl();
		
	}
	
	
		
	@Override
	protected IDataSet getDataSet() throws Exception {
		return new FlatXmlDataSet(this.getClass().getResourceAsStream ("TestDataSeed.xml"));
	}
	
	protected DatabaseOperation getSetUpOperation() 
	  throws Exception {
		return DatabaseOperation.CLEAN_INSERT;
	 }
	
	protected DatabaseOperation getTearDownOperation() 
	  throws Exception {
		return DatabaseOperation.NONE;
	}
	
	public void testLoadCustomer() {
		AbstractCustomer customer;
		try {
			customer = customerOrderServices.loadCustomer(customerId);
			assertNotNull(customer);
			assertEquals(customerId,customer.getCustomerId() );
			assertNull(customer.getOpenOrder());
			customerOrderServices.openOrder(customerId);
			customer = customerOrderServices.loadCustomer(customerId);
			assertNotNull(customer.getOpenOrder());
			customerOrderServices.addLineItem(customerId, 1, 1);
			customer = customerOrderServices.loadCustomer(customerId);
			assertNotNull(customer.getOpenOrder());
			assertNotNull(customer.getOpenOrder().getLineitems());
			assertTrue(customer.getOpenOrder().getLineitems().size() > 0);
			customerOrderServices.submit(customerId);
			customer = customerOrderServices.loadCustomer(customerId);
			assertNull(customer.getOpenOrder());
			
			//Inherit Test
			assertTrue(customer instanceof ResidentialCustomer );
			
			customer = customerOrderServices.loadCustomer(businessCustomerId);
			assertTrue(customer instanceof BusinessCustomer);
			
		} catch (CustomerDoesNotExistException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			fail("Customer Does Not Exist");
		} catch (GeneralPersistenceException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			fail("Persistence Error");
		} catch (OrderAlreadyOpenException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			fail("Order Already Open");
		} catch (OrderNotOpenException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			fail("Order Not Open");
		} catch (ProductDoesNotExistException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			fail("No Such Product");
		} catch (InvalidQuantityException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			fail("InvalidQuantity");
		} catch (NoLineItemsException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			fail("No Line Item");
		}
		
	}
	
	public void testLoadCustomerFail() {
		AbstractCustomer customer;
		try {
			customer = customerOrderServices.loadCustomer(2222);
			if(customer == null) assert true;
			else
			fail("This Customer should not be here" );
		} catch (CustomerDoesNotExistException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			assertTrue("This is the correct Exception", true);
		} catch (GeneralPersistenceException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			fail("Persistence Error");
		}
		
	}
	

	public void testOpenOrder() {
		try {
			Order openOrder = customerOrderServices.openOrder(customerId);
			assertNotNull(openOrder);
			AbstractCustomer customer = customerOrderServices.loadCustomer(customerId);
			assertNotNull(customer.getOpenOrder());
			assertEquals(customer.getOpenOrder().getOrderId(), openOrder.getOrderId());
			try
			{
			customerOrderServices.openOrder(customerId);
			}
			catch (OrderAlreadyOpenException e) {
				// TODO Auto-generated catch block
				assertTrue("This is the Correct Exception",true);
			}
			customerOrderServices.addLineItem(customerId, 1, 1);
			customerOrderServices.submit(customerId);
		} catch (OrderAlreadyOpenException e) {
			fail("Order Already Open");
		} catch (CustomerDoesNotExistException e) {
			fail("Customer does not exist.");
		} 
		catch (GeneralPersistenceException e) {
			fail("General Persistence Error");
		} 
		catch (OrderNotOpenException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			fail("Order Not Open");
		} catch (ProductDoesNotExistException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			fail("No Such Product");
		} catch (InvalidQuantityException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			fail("InvalidQuantity");
		} catch (NoLineItemsException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			fail("No Line Item");
		}
	}
	
	public void testAddLineItem() {
		try {
			Order openOrder = customerOrderServices.openOrder(customerId);
			LineItem lineItem = customerOrderServices.addLineItem(customerId, 1, 1);
			assertNotNull(lineItem);
			AbstractCustomer customer = customerOrderServices.loadCustomer(customerId);
			Set lineItems = customer.getOpenOrder().getLineitems();
			assertEquals(1,lineItems.size());
			LineItem li = ((LineItem)(lineItems.iterator().next()));
			assertEquals(li.getProductId(),lineItem.getProductId());
			assertEquals(li.getOrderId(),lineItem.getOrderId());
			
			//Add Line Item for Same Product, should just bump Quantity
			customerOrderServices.addLineItem(customerId, 1, 1);
			customer = customerOrderServices.loadCustomer(customerId);
			lineItems = customer.getOpenOrder().getLineitems();
			assertEquals(lineItems.size(), 1);
			
			//Add a Second Line Item ...
			customerOrderServices.addLineItem(customerId, 2, 1);
			customer = customerOrderServices.loadCustomer(customerId);
			lineItems = customer.getOpenOrder().getLineitems();
			assertEquals(lineItems.size(), 2);
			
			try
			{
				customerOrderServices.addLineItem(customerId, 1, 0);
			}
			catch (InvalidQuantityException e) {
				assertTrue("Correct Error, cannot add 0 items",true);
			} 
			try
			{
				customerOrderServices.addLineItem(customerId, 1, -1);
			}
			catch (InvalidQuantityException e) {
				assertTrue("Correct Error, cannot add negative items",true);
			} 
			
			try
			{
				customerOrderServices.addLineItem(customerId, 444, 1);
				fail("Cannot add product that does not exist");
			}
			catch (ProductDoesNotExistException e) {
				assertTrue("Correct Exception",true);
			}
			customerOrderServices.submit(customerId);
			
		} catch (OrderAlreadyOpenException e) {
			// TODO Auto-generated catch block
			fail("Order Already Open");
		} catch (CustomerDoesNotExistException e) {
			// TODO Auto-generated catch block
			fail("Customer does not exist.");
		} 
		catch (GeneralPersistenceException e) {
			fail("General Persistence Error");
		} catch (OrderNotOpenException e) {
			fail("Order Not Open");
		} catch (ProductDoesNotExistException e) {
			fail("No Such Product");
		} catch (InvalidQuantityException e) {
			fail("Invlid Quantity");
		} catch (NoLineItemsException e) {
			fail("No Line Items");
		}
	}

	public void testSubmit() {
		Order openOrder;
		try {
			try {
				customerOrderServices.submit(customerId);
				fail("Non Open Order Got Submitted");
			} catch (OrderNotOpenException e) {
				assertTrue("Correct Exception, No Open Order",true);
			} catch (NoLineItemsException e) {
				fail("Not correct Exception");
			}
			openOrder = customerOrderServices.openOrder(customerId);
			assertNotNull(openOrder);
			
			try {
				customerOrderServices.submit(customerId);
				fail("Order with No LineItems was submitted");
			} catch (OrderNotOpenException e) {
				
				fail("Not correct Exception");
			} catch (NoLineItemsException e) {
				assertTrue("Correct Exception, No LineItems",true);
			}
			
			customerOrderServices.addLineItem(customerId, 1, 1);
			customerOrderServices.submit(customerId);
			AbstractCustomer customer = customerOrderServices.loadCustomer(customerId);
			//Open Order Should be Null
			assertNull(customer.getOpenOrder());
			 
		} catch (CustomerDoesNotExistException e) {
			fail("Customer Does Not Exist");
		} catch (OrderAlreadyOpenException e) {
			fail("No Open Order");
		} catch (GeneralPersistenceException e) {
			fail("Persistence Failure");
		} catch (OrderNotOpenException e) {
			fail("Order Already Open");
		} catch (ProductDoesNotExistException e) {
			fail("No Such Product");
		} catch (InvalidQuantityException e) {
			fail("Not a Valid Quantity");
		} catch (NoLineItemsException e) {
			fail("No Line Items");
		}
			
		
	}

	
	public void testRemoveLineItem() {
			try {
				
				try
				{
					customerOrderServices.removeLineItem(2, 1);
					fail("Cannot remove Line from non open Order");
				}
				catch(OrderNotOpenException e)
				{
					assertTrue("Correct, Order Not Open",true);
				}
				Order openOrder = customerOrderServices.openOrder(customerId);
				
				try
				{
					customerOrderServices.removeLineItem(2, 1);
					assertTrue("Will let pass to avoid extra query",true);
				}
				catch(NoLineItemsException e)
				{
					assertTrue("Correct, Order Not Open",true);
				}
				LineItem lineItem = customerOrderServices.addLineItem(customerId, 1, 1);
				customerOrderServices.removeLineItem(2, 1);
				
				AbstractCustomer customer = customerOrderServices.loadCustomer(customerId);
				Set lineItems = customer.getOpenOrder().getLineitems();
				assertEquals(lineItems.size(), 0);
				
				try
				{
					customerOrderServices.removeLineItem(2, 333);
					fail("Cannot remove product that does not exist");
				}
				catch (ProductDoesNotExistException e) {
					assertTrue("Correct Exception",true);
				}
				customerOrderServices.addLineItem(customerId, 1, 1);
				customerOrderServices.submit(customerId);
				
			} catch (CustomerDoesNotExistException e) {
				fail("Customer Does Not Exist");
			} catch (OrderAlreadyOpenException e) {
				fail("Order Already Open");
			} catch (GeneralPersistenceException e) {
				fail("General Persistence");
			} catch (OrderNotOpenException e) {
				fail("Order Not OPen");
			} catch (ProductDoesNotExistException e) {
				fail("No Such Product");
			} catch (InvalidQuantityException e) {
				fail("Invalid Quantity");
			} catch (NoLineItemsException e) {
				fail("No Line Item");
			}
			
	}

	

}
